@@ -1,7 +1,6 @@
-/* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2009-2018 Brazil
-  Copyright(C) 2018-2020 Sutou Kouhei <kou@clear-code.com>
+  Copyright(C) 2009-2018  Brazil
+  Copyright(C) 2018-2021  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -59,21 +58,33 @@ extern "C" {
 
 #define DB_OBJ(obj) ((grn_db_obj *)obj)
 
+#define GRN_DEFINE_NAME(obj)                                            \
+  const char *name;                                                     \
+  char name_buffer[GRN_TABLE_MAX_KEY_SIZE];                             \
+  int name_size;                                                        \
+  do {                                                                  \
+    if (DB_OBJ(obj)->id == GRN_ID_NIL) {                                \
+      name = "(temporary)";                                             \
+      name_size = strlen(name);                                         \
+    } else {                                                            \
+      name_size = grn_obj_name(ctx, (grn_obj *)obj,                     \
+                               name_buffer, GRN_TABLE_MAX_KEY_SIZE);    \
+      name = name_buffer;                                               \
+      if (name_size == 0) {                                             \
+        name = "(anonymous)";                                           \
+        name_size = strlen(name);                                       \
+      }                                                                 \
+    }                                                                   \
+  } while (false)
+
 extern bool grn_enable_reference_count;
 
 typedef struct _grn_db grn_db;
 typedef struct _grn_proc grn_proc;
 
-typedef enum {
-  GRN_DEFERRED_UNREF_RECURSIVE_NO,
-  GRN_DEFERRED_UNREF_RECURSIVE_YES,
-  GRN_DEFERRED_UNREF_RECURSIVE_DEPENDENT,
-} grn_deferred_unref_recursive;
-
 typedef struct {
   uint32_t count;
-  grn_id id;
-  grn_deferred_unref_recursive recursive;
+  grn_obj ids;
 } grn_deferred_unref;
 
 struct _grn_db {
@@ -193,6 +204,15 @@ grn_db_get_option_values(grn_ctx *ctx,
 grn_rc grn_db_clear_option_values(grn_ctx *ctx,
                                   grn_obj *db,
                                   grn_id id);
+grn_rc grn_obj_refer_auto_release(grn_ctx *ctx,
+                                  grn_obj *obj,
+                                  uint32_t count);
+grn_rc grn_obj_refer_recursive_auto_release(grn_ctx *ctx,
+                                            grn_obj *obj,
+                                            uint32_t count);
+grn_rc grn_obj_refer_recursive_dependent_auto_release(grn_ctx *ctx,
+                                                      grn_obj *obj,
+                                                      uint32_t count);
 grn_rc
 grn_db_add_deferred_unref(grn_ctx *ctx,
                           grn_obj *db,
@@ -210,6 +230,9 @@ grn_rc _grn_table_delete_by_id(grn_ctx *ctx, grn_obj *table, grn_id id,
 
 grn_id grn_table_get_v(grn_ctx *ctx, grn_obj *table, const void *key, int key_size,
                        void **value);
+grn_id grn_table_get_by_key(grn_ctx *ctx,
+                            grn_obj *table,
+                            grn_obj *key);
 grn_id grn_table_add_v(grn_ctx *ctx, grn_obj *table, const void *key, int key_size,
                        void **value, int *added);
 grn_id grn_table_add_by_key(grn_ctx *ctx,
