@@ -1,5 +1,6 @@
 /*
-  Copyright(C) 2018 Brazil
+  Copyright(C) 2018  Brazil
+  Copyright(C) 2021  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -56,6 +57,9 @@ grn_rc grn_options_clear(grn_ctx *ctx,
                          grn_options *options,
                          grn_id id);
 
+grn_rc grn_options_warm(grn_ctx *ctx,
+                        grn_options *options);
+
 #ifdef __cplusplus
 }
 #endif
