@@ -1,4 +1,3 @@
-/* -*- c-basic-offset: 2 -*- */
 /*
   Copyright(C) 2009-2018  Brazil
   Copyright(C) 2019-2021  Sutou Kouhei <kou@clear-code.com>
@@ -63,6 +62,7 @@ void grn_proc_init_object_inspect(grn_ctx *ctx);
 void grn_proc_init_object_list(grn_ctx *ctx);
 void grn_proc_init_object_remove(grn_ctx *ctx);
 void grn_proc_init_object_set_visibility(grn_ctx *ctx);
+void grn_proc_init_object_warm(grn_ctx *ctx);
 void grn_proc_init_query(grn_ctx *ctx);
 void grn_proc_init_query_parallel_or(grn_ctx *ctx);
 void grn_proc_init_query_expand(grn_ctx *ctx);
