@@ -1,7 +1,6 @@
-/* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2009-2018 Brazil
-  Copyright(C) 2018-2020 Sutou Kouhei <kou@clear-code.com>
+  Copyright(C) 2009-2018  Brazil
+  Copyright(C) 2018-2021  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -26,6 +25,7 @@
 #include "grn_expr_executor.h"
 #include "grn_group.h"
 #include "grn_output_columns.h"
+#include "grn_sort.h"
 #include "grn_table.h"
 #include "grn_vector.h"
 
@@ -924,7 +924,7 @@ grn_table_group_single_key_records(grn_ctx *ctx, grn_obj *table,
         cursor,
         grn_table_group_single_key_records_foreach_fix_size,
         &data);
-      GRN_RA_CACHE_FIN((grn_ra *)(data.key.object), &(data.cache));
+      GRN_RA_CACHE_FIN(ctx, (grn_ra *)(data.key.object), &(data.cache));
     } else if (data.key.object->header.type == GRN_COLUMN_VAR_SIZE &&
                data.key.is_reference_column) {
       grn_table_cursor_foreach(
@@ -1040,7 +1040,7 @@ grn_table_group_with_range_gap(grn_ctx *ctx, grn_obj *table,
                 }
               }
             }
-            GRN_RA_CACHE_FIN(ra, &cache);
+            GRN_RA_CACHE_FIN(ctx, ra, &cache);
           }
           break;
         case GRN_COLUMN_VAR_SIZE :
@@ -1385,99 +1385,143 @@ grn_table_group_update_aggregated_value_type_id(grn_ctx *ctx,
   }
 }
 
+static grn_rc
+grn_table_group_results_prepare(grn_ctx *ctx,
+                                grn_table_group_result *results,
+                                int n_results,
+                                grn_obj *table,
+                                grn_table_sort_key *keys,
+                                int n_keys)
+{
+  int i;
+  grn_table_group_result *rp;
+  for (i = 0, rp = results; i < n_results; i++, rp++) {
+    if (!rp->table) {
+      grn_table_flags flags;
+      grn_obj *key_type = NULL;
+      uint32_t additional_value_size;
+
+      flags = GRN_OBJ_TABLE_HASH_KEY|
+        GRN_OBJ_WITH_SUBREC|
+        GRN_OBJ_UNIT_USERDEF_DOCUMENT;
+      if (n_keys == 0 && n_results == 1) {
+        key_type = grn_ctx_at(ctx, GRN_DB_SHORT_TEXT);
+      } else if (n_keys == 1) {
+        grn_id key_type_id = grn_obj_get_range(ctx, keys[0].key);
+        if (key_type_id == GRN_ID_NIL) {
+          grn_obj inspected;
+          GRN_TEXT_INIT(&inspected, 0);
+          grn_inspect(ctx, &inspected, keys[0].key);
+          ERR(GRN_INVALID_ARGUMENT,
+              "[table][group] unknown key type: %.*s",
+              (int)GRN_TEXT_LEN(&inspected),
+              GRN_TEXT_VALUE(&inspected));
+          GRN_OBJ_FIN(ctx, &inspected);
+          return ctx->rc;
+        }
+        key_type = grn_ctx_at(ctx, key_type_id);
+      } else {
+        flags |= GRN_OBJ_KEY_VAR_SIZE;
+      }
+      additional_value_size = grn_rset_recinfo_calc_values_size(ctx,
+                                                                rp->flags);
+      rp->table = grn_table_create_with_max_n_subrecs(ctx, NULL, 0, NULL,
+                                                      flags,
+                                                      key_type, table,
+                                                      rp->max_n_subrecs,
+                                                      additional_value_size);
+      if (key_type) {
+        grn_obj_unref(ctx, key_type);
+      }
+      if (!rp->table) {
+        return ctx->rc;
+      }
+    }
+    DB_OBJ(rp->table)->group.flags = rp->flags;
+    DB_OBJ(rp->table)->group.calc_target = rp->calc_target;
+    grn_table_group_update_aggregated_value_type_id(ctx, rp->table);
+    if (rp->flags & GRN_TABLE_GROUP_CALC_AGGREGATOR) {
+      DB_OBJ(rp->table)->group.aggregators = rp->aggregators;
+      DB_OBJ(rp->table)->group.n_aggregators = rp->n_aggregators;
+    } else {
+      DB_OBJ(rp->table)->group.aggregators = NULL;
+      DB_OBJ(rp->table)->group.n_aggregators = 0;
+    }
+    {
+      uint32_t j;
+      for (j = 0; j < DB_OBJ(rp->table)->group.n_aggregators; j++) {
+        grn_table_group_aggregator *aggregator =
+          DB_OBJ(rp->table)->group.aggregators[j];
+        grn_table_group_aggregator_prepare(ctx, aggregator, rp->table);
+        if (ctx->rc != GRN_SUCCESS) {
+          return ctx->rc;
+        }
+      }
+    }
+  }
+  return ctx->rc;
+}
+
+static grn_rc
+grn_table_group_results_postpare(grn_ctx *ctx,
+                                 grn_table_group_result *results,
+                                 int n_results)
+{
+  int i;
+  grn_table_group_result *rp;
+  for (i = 0, rp = results; i < n_results; i++, rp++) {
+    GRN_TABLE_GROUPED_ON(rp->table);
+    DB_OBJ(rp->table)->group.calc_target = NULL;
+    {
+      uint32_t j;
+      for (j = 0; j < DB_OBJ(rp->table)->group.n_aggregators; j++) {
+        grn_table_group_aggregator *aggregator =
+          DB_OBJ(rp->table)->group.aggregators[j];
+        grn_table_group_aggregator_postpare(ctx, aggregator);
+      }
+    }
+  }
+  return ctx->rc;
+}
+
 grn_rc
 grn_table_group(grn_ctx *ctx, grn_obj *table,
                 grn_table_sort_key *keys, int n_keys,
                 grn_table_group_result *results, int n_results)
 {
-  grn_rc rc = GRN_SUCCESS;
-  bool group_by_all_records = false;
-  if (n_keys == 0 && n_results == 1) {
-    group_by_all_records = GRN_TRUE;
-  } else if (!table) {
-    ERR(GRN_INVALID_ARGUMENT, "[table][group] table is NULL");
-    return GRN_INVALID_ARGUMENT;
-  } else if (n_keys == 0) {
-    ERR(GRN_INVALID_ARGUMENT, "[table][group] no keys");
-    return GRN_INVALID_ARGUMENT;
-  } else if (n_results == 0) {
-    ERR(GRN_INVALID_ARGUMENT, "[table][group] no results");
-    return GRN_INVALID_ARGUMENT;
-  }
   GRN_API_ENTER;
+  if (!table) {
+    ERR(GRN_INVALID_ARGUMENT, "[table][group] table is NULL");
+    goto exit;
+  }
+  if (n_results == 0) {
+    ERR(GRN_INVALID_ARGUMENT, "[table][group] no results");
+    goto exit;
+  }
+  bool group_by_all_records = false;
+  if (n_keys == 0) {
+    if (n_results == 1) {
+      group_by_all_records = true;
+    } else {
+      ERR(GRN_INVALID_ARGUMENT, "[table][group] no keys");
+      goto exit;
+    }
+  }
   {
-    int k, r;
+    int k;
     grn_table_sort_key *kp;
-    grn_table_group_result *rp;
     for (k = 0, kp = keys; k < n_keys; k++, kp++) {
       if ((kp->flags & GRN_TABLE_GROUP_BY_COLUMN_VALUE) && !kp->key) {
         ERR(GRN_INVALID_ARGUMENT, "[table][group] column is missing: %d", k);
         goto exit;
       }
     }
-    for (r = 0, rp = results; r < n_results; r++, rp++) {
-      if (!rp->table) {
-        grn_table_flags flags;
-        grn_obj *key_type = NULL;
-        uint32_t additional_value_size;
-
-        flags = GRN_OBJ_TABLE_HASH_KEY|
-          GRN_OBJ_WITH_SUBREC|
-          GRN_OBJ_UNIT_USERDEF_DOCUMENT;
-        if (group_by_all_records) {
-          key_type = grn_ctx_at(ctx, GRN_DB_SHORT_TEXT);
-        } else if (n_keys == 1) {
-          grn_id key_type_id = grn_obj_get_range(ctx, keys[0].key);
-          if (key_type_id == GRN_ID_NIL) {
-            grn_obj inspected;
-            GRN_TEXT_INIT(&inspected, 0);
-            grn_inspect(ctx, &inspected, keys[0].key);
-            ERR(GRN_INVALID_ARGUMENT,
-                "[table][group] unknown key type: %.*s",
-                (int)GRN_TEXT_LEN(&inspected),
-                GRN_TEXT_VALUE(&inspected));
-            GRN_OBJ_FIN(ctx, &inspected);
-            goto exit;
-          }
-          key_type = grn_ctx_at(ctx, key_type_id);
-        } else {
-          flags |= GRN_OBJ_KEY_VAR_SIZE;
-        }
-        additional_value_size = grn_rset_recinfo_calc_values_size(ctx,
-                                                                  rp->flags);
-        rp->table = grn_table_create_with_max_n_subrecs(ctx, NULL, 0, NULL,
-                                                        flags,
-                                                        key_type, table,
-                                                        rp->max_n_subrecs,
-                                                        additional_value_size);
-        if (key_type) {
-          grn_obj_unref(ctx, key_type);
-        }
-        if (!rp->table) {
-          goto exit;
-        }
-      }
-      DB_OBJ(rp->table)->group.flags = rp->flags;
-      DB_OBJ(rp->table)->group.calc_target = rp->calc_target;
-      grn_table_group_update_aggregated_value_type_id(ctx, rp->table);
-      if (rp->flags & GRN_TABLE_GROUP_CALC_AGGREGATOR) {
-        DB_OBJ(rp->table)->group.aggregators = rp->aggregators;
-        DB_OBJ(rp->table)->group.n_aggregators = rp->n_aggregators;
-      } else {
-        DB_OBJ(rp->table)->group.aggregators = NULL;
-        DB_OBJ(rp->table)->group.n_aggregators = 0;
-      }
-      {
-        uint32_t i;
-        for (i = 0; i < DB_OBJ(rp->table)->group.n_aggregators; i++) {
-          grn_table_group_aggregator *aggregator =
-            DB_OBJ(rp->table)->group.aggregators[i];
-          grn_table_group_aggregator_prepare(ctx, aggregator, rp->table);
-          if (ctx->rc != GRN_SUCCESS) {
-            goto exit;
-          }
-        }
-      }
+    grn_table_group_results_prepare(ctx,
+                                    results, n_results,
+                                    table,
+                                    keys, n_keys);
+    if (ctx->rc != GRN_SUCCESS) {
+      goto exit;
     }
     if (group_by_all_records) {
       grn_table_group_all_records(ctx, table, results);
@@ -1541,48 +1585,10 @@ grn_table_group(grn_ctx *ctx, grn_obj *table,
       }
       GRN_FREE(data.keys);
     }
-    for (r = 0, rp = results; r < n_results; r++, rp++) {
-      GRN_TABLE_GROUPED_ON(rp->table);
-      DB_OBJ(rp->table)->group.calc_target = NULL;
-      {
-        uint32_t i;
-        for (i = 0; i < DB_OBJ(rp->table)->group.n_aggregators; i++) {
-          grn_table_group_aggregator *aggregator =
-            DB_OBJ(rp->table)->group.aggregators[i];
-          grn_table_group_aggregator_postpare(ctx, aggregator);
-        }
-      }
-    }
+    grn_table_group_results_postpare(ctx, results, n_results);
   }
 exit :
-  GRN_API_RETURN(rc);
-}
-
-static void
-grn_table_group_keys_parse_one(grn_ctx *ctx,
-                               grn_obj *table,
-                               grn_obj *keys_expression,
-                               uint32_t code_start_offset,
-                               uint32_t code_end_offset,
-                               grn_table_sort_key *key)
-{
-  grn_expr *expr = (grn_expr *)keys_expression;
-  if ((code_end_offset - code_start_offset) == 1 &&
-      expr->codes[code_start_offset].op == GRN_OP_GET_VALUE &&
-      expr->codes[code_start_offset].value) {
-    key->key = expr->codes[code_start_offset].value;
-    if (grn_obj_is_accessor(ctx, key->key)) {
-      key->key = grn_accessor_copy(ctx, key->key);
-    } else {
-      grn_obj_refer(ctx, key->key);
-    }
-  } else {
-    grn_obj *sliced_expr = grn_expr_slice(ctx,
-                                          keys_expression,
-                                          code_start_offset,
-                                          code_end_offset);
-    key->key = sliced_expr;
-  }
+  GRN_API_RETURN(ctx->rc);
 }
 
 grn_table_sort_key *
@@ -1592,94 +1598,11 @@ grn_table_group_keys_parse(grn_ctx *ctx,
                            int32_t raw_keys_length,
                            int *n_keys)
 {
-  GRN_API_ENTER;
-  if (raw_keys_length < 0) {
-    raw_keys_length = strlen(raw_keys);
-  }
-  grn_table_sort_key *keys = NULL;
-  if (grn_expr_is_v1_columns_format(ctx, raw_keys, raw_keys_length)) {
-    keys = grn_table_sort_key_from_str(ctx,
+  return
+    grn_table_sort_keys_parse_internal(ctx,
+                                       table,
                                        raw_keys,
                                        raw_keys_length,
-                                       table,
-                                       n_keys);
-    GRN_API_RETURN(keys);
-  }
-
-  grn_obj offsets;
-  GRN_UINT32_INIT(&offsets, GRN_OBJ_VECTOR);
-  grn_obj *keys_expression = NULL;
-  grn_obj *variable;
-  GRN_EXPR_CREATE_FOR_QUERY(ctx, table, keys_expression, variable);
-  if (ctx->rc != GRN_SUCCESS) {
-    goto exit;
-  }
-  grn_expr_parse(ctx,
-                 keys_expression,
-                 raw_keys,
-                 raw_keys_length,
-                 NULL,
-                 GRN_OP_MATCH,
-                 GRN_OP_AND,
-                 GRN_EXPR_SYNTAX_GROUP_KEYS);
-  if (ctx->rc != GRN_SUCCESS) {
-    goto exit;
-  }
-
-  grn_output_columns_get_offsets(ctx, keys_expression, &offsets);
-  if (ctx->rc != GRN_SUCCESS) {
-    goto exit;
-  }
-
-  size_t n_offsets = GRN_UINT32_VECTOR_SIZE(&offsets) / 2;
-  keys = GRN_MALLOCN(grn_table_sort_key, n_offsets);
-  if (!keys) {
-    goto exit;
-  }
-  size_t i;
-  for (i = 0; i < n_offsets; i++) {
-    keys[i].key = NULL;
-  }
-
-  for (i = 0; i < n_offsets; i++) {
-    uint32_t code_start_offset;
-    uint32_t code_end_offset;
-    code_start_offset = GRN_UINT32_VALUE_AT(&offsets, i * 2);
-    code_end_offset = GRN_UINT32_VALUE_AT(&offsets, i * 2 + 1);
-    grn_table_group_keys_parse_one(ctx,
-                                   table,
-                                   keys_expression,
-                                   code_start_offset,
-                                   code_end_offset,
-                                   &(keys[i]));
-    if (ctx->rc != GRN_SUCCESS) {
-      goto exit;
-    }
-  }
-  *n_keys = n_offsets;
-
-exit :
-  if (ctx->rc != GRN_SUCCESS) {
-    if (keys) {
-      size_t i;
-      size_t n_offsets = GRN_UINT32_VECTOR_SIZE(&offsets) / 2;
-      for (i = 0; i < n_offsets; i++) {
-        if (keys[i].key) {
-          if (grn_obj_is_accessor(ctx, keys[i].key)) {
-            grn_obj_close(ctx, keys[i].key);
-          } else {
-            grn_obj_unref(ctx, keys[i].key);
-          }
-        }
-      }
-      GRN_FREE(keys);
-      keys = NULL;
-    }
-  }
-  if (keys_expression) {
-    grn_obj_close(ctx, keys_expression);
-  }
-  GRN_OBJ_FIN(ctx, &offsets);
-
-  GRN_API_RETURN(keys);
+                                       n_keys,
+                                       GRN_TABLE_SORT_KEYS_PARSE_MODE_GROUP);
 }
